# This file was automatically generated.
# Do not edit, you'll loose your changes anyway.
package Prima::Config;
use vars qw(%Config %Config_inst);

%Config_inst = (
	incpaths              => [ '$(lib)/Prima/CORE','$(lib)/Prima/CORE/generic','C:/usr/local/perl/sb32p.522.01/c/lib/pkgconfig/../../include/libpng16' ],
	gencls                => '$(bin)/gencls.bat',
	tmlink                => '$(bin)/tmlink.bat',
	libname               => '$(lib)/auto/Prima/libPrima.a',
	dlname                => '$(lib)/auto/Prima/Prima.xs.dll',
	ldpaths               => [],

	inc                   => '-I$(lib)/Prima/CORE -I$(lib)/Prima/CORE/generic -IC:/usr/local/perl/sb32p.522.01/c/lib/pkgconfig/../../include/libpng16',
	libs                  => '$(lib)/auto/Prima/libPrima.a',
);

%Config = (
	ifs                   => '\/',
	quote                 => '\"',
	platform              => 'win32',
	incpaths              => [ 'C:/home/Prima/Prima/include','C:/home/Prima/Prima/include/generic','C:/usr/local/perl/sb32p.522.01/c/lib/pkgconfig/../../include/libpng16' ],
	gencls                => 'C:/home/Prima/Prima/blib/script/gencls.bat',
	tmlink                => 'C:/home/Prima/Prima/blib/script/tmlink.bat',
	scriptext             => '.bat',
	genclsoptions         => '--tml --h --inc',
	cobjflag              => '-o ',
	coutexecflag          => '-o ',
	clinkprefix           => '',
	clibpathflag          => '-L',
	cdefs                 => [],
	libext                => '.a',
	libprefix             => 'lib',
	libname               => 'C:/home/Prima/Prima/blib/arch/auto/Prima/libPrima.a',
	dlname                => 'C:/home/Prima/Prima/blib/arch/auto/Prima/Prima.xs.dll',
	ldoutflag             => '-o ',
	ldlibflag             => '-l',
	ldlibpathflag         => '-L',
	ldpaths               => [],
	ldlibs                => ['gif','jpeg','gomp','gdi32','mpr','winspool','comdlg32','msimg32','png16','tiff','Xpm'],
	ldlibext              => '',
	inline                => 'inline',
	dl_load_flags         => 0,
	optimize              => '-s -O2',
	openmp                => '-fopenmp',

	inc                   => '-IC:/home/Prima/Prima/include -IC:/home/Prima/Prima/include/generic -IC:/usr/local/perl/sb32p.522.01/c/lib/pkgconfig/../../include/libpng16',
	define                => '',
	libs                  => 'C:/home/Prima/Prima/blib/arch/auto/Prima/libPrima.a',
);

1;
